/* Copyright (c) 2014 Nokia Solutions and Networks
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/**
 * CPLD SPI driver for the NSN FZM
 *
 * Provides read/write primitives to access CPLD registers, FPGA registers, and
 * FPGA configuration over SPI via the CPLD.
 *
 * CPLD registers are 8-bit, CPLD addresses are 10 bits.
 * FPGA registers are 32-bit, FPGA addresses are 18 bits.
 *
 * Protocol for CPLD/FPGA/FPGA Config Access
 * - Series of 8-bit words
 * - First word specifies target device, read/write operation, high addr bits
 * - Following words specify more address bits or data depending on the target
 *   device.
 *
 * FIRST WORD
 * ---------------------------------------------
 * | S3 | S2 | S1 | S0 | FILL | RW | AH1 | AH2 |
 * ---------------------------------------------
 *
 * <S3..S0>     = 0001 for CPLD access.
 *                0010 for FPGA configuration.
 *                0011 for FPGA access.
 * FILL         = Don't care
 * RW           = 1 for read and 0 for write access. (Don't care for FPGA cfg)
 * <AH1..AH0>   = high address bits, set correspondingly to:
 *                bits <9:8> in the 10-bit CPLD address.
 *                bits <17:16> in the 18-bit FPGA address.
 *                Don't care for FPGA config.
 *
 * SECOND WORD
 * -------------------------------------------------
 * | AD7 | AD6 | AD5 | AD4 | AD3 | AD2 | AD1 | AD0 |
 * -------------------------------------------------
 *
 * <AD7..AD0>   = for CPLD access, bits <7:0> in 10-bit CPLD address.
 *              = for FPGA access, bits <15:8> in 18-bit FPGA address.
 *              = for FPGA config, 1st data byte of data config stream.
 *
 * THIRD WORD
 * -------------------------------------------------
 * | AD7 | AD6 | AD5 | AD4 | AD3 | AD2 | AD1 | AD0 |
 * -------------------------------------------------
 *
 * <AD7..AD0>   = for CPLD access, 1st data byte.
 *              = for FPGA access, bits <7:0> in 18-bit FPGA address.
 *              = for FPGA config, 2nd data byte of data config stream.
 *
 * Notes on data bytes:
 * - For CPLD, if more than 1 data byte is supplied in an access, the
 *   address shall auto-increment to the next 32-bit boundary.  This shall
 *   apply to both read and write accesses.
 *
 * - For FPGA, if more than 4 data bytes are supplied in an access, the
 *   address shall auto-increment to the next 32-bit boundary.  This shall
 *   apply to both read and write accesses.  Data should be accessed in
 *   multiples of four bytes to insure proper transmission of data.
 *
 * - Chip select deassertion terminates the access.
 */


#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/spi/spi.h>
#include <linux/hardirq.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/of.h>
#include <stdbool.h>
#include "fzm_cpldspi.h"
#include "cpld_internal.h"

static struct spi_device *spidev;

static DEFINE_MUTEX(mutex_cpld);
static DEFINE_MUTEX(mutex_fpga);

#define FZM_CPLDSPI_DEV_SHIFT            4
#define FZM_CPLDSPI_OP_READ              4

#define FZM_CPLDSPI_CPLD_HIADDR_SHIFT    8
#define FZM_CPLDSPI_CPLD_HIADDR_MASK     0x00000003
#define FZM_CPLDSPI_CPLD_LOADDR_MASK     0x000000FC

#define FZM_CPLDSPI_FPGA_HIADDR_SHIFT    16
#define FZM_CPLDSPI_FPGA_HIADDR_MASK     FZM_CPLDSPI_CPLD_HIADDR_MASK
#define FZM_CPLDSPI_FPGA_MIDADDR_MASK    0x000000FF
#define FZM_CPLDSPI_FPGA_MIDADDR_SHIFT   8
#define FZM_CPLDSPI_FPGA_LOADDR_MASK     FZM_CPLDSPI_CPLD_LOADDR_MASK


void fzm_cpldspi_lock(fzm_cpldspi_target_dev_t dev)
{
   switch (dev) {
   case FZM_CPLDSPI_CPLD:
      mutex_lock(&mutex_cpld);
      break;
   case FZM_CPLDSPI_FPGA_CONFIG:
      mutex_lock(&mutex_fpga);
      break;
   }
}
EXPORT_SYMBOL(fzm_cpldspi_lock);


int fzm_cpldspi_lock_interruptible(fzm_cpldspi_target_dev_t dev)
{
   switch (dev) {
   case FZM_CPLDSPI_CPLD:
      return mutex_lock_interruptible(&mutex_cpld);
   case FZM_CPLDSPI_FPGA_CONFIG:
      return mutex_lock_interruptible(&mutex_fpga);
   }

   return -EINVAL;
}
EXPORT_SYMBOL(fzm_cpldspi_lock_interruptible);


int fzm_cpldspi_trylock(fzm_cpldspi_target_dev_t dev)
{
   switch (dev) {
   case FZM_CPLDSPI_CPLD:
      return mutex_trylock(&mutex_cpld);
   case FZM_CPLDSPI_FPGA_CONFIG:
      return mutex_trylock(&mutex_fpga);
   }

   return 0; /* unsuccessful */
}
EXPORT_SYMBOL(fzm_cpldspi_trylock);


void fzm_cpldspi_unlock(fzm_cpldspi_target_dev_t dev)
{
   switch (dev) {
   case FZM_CPLDSPI_CPLD:
      mutex_unlock(&mutex_cpld);
      break;
   case FZM_CPLDSPI_FPGA_CONFIG:
      mutex_unlock(&mutex_fpga);
      break;
   }
}
EXPORT_SYMBOL(fzm_cpldspi_unlock);


/**
 * fzm_cpldspi_rw
 * @dev: target device (CPLD, FPGA_CONFIG, or FPGA)
 * @op: FZM_CPLDSPI_READ or FZM_CPLDSPI_WRITE.
 *      Ignored if dev=FZM_CPLDSPI_FPGA_CONFIG.
 * @addr: register address on the target device.
 *      Ignored if dev=FZM_CPLDSPI_FPGA_CONFIG.
 * @buf: pointer to buffer of one or more bytes to be read or written.
 * @len: length of buffer in bytes.  This will also be the number of bytes read
 *       or written.  The len must be a multiple of 4 for FPGA accesses.
 *
 * Cannot be called from interrupt context.
 *
 * Returns 0 upon success, negative errno value upon failure.
 */
static int fzm_cpldspi_rw(fzm_cpldspi_target_dev_t dev,
                          unsigned addr, void *buf, size_t len, bool read)
{
   struct spi_transfer t[2];
   struct spi_message m;
   int ret = 0;
   u8 cmd[3];

   /* printk("<1> SPI RW!!!!!!\n");
      printk("<1> dev %u r %u addr %u len %u\n",dev, read, addr, len); */

   /* spi_sync below cannot be called from interrupt context */
   WARN_ON(!preemptible());

   if (WARN(!spidev,
            "cpldspi: spi driver not ready. dev %u r %u addr %u len %u\n",
            dev, read, addr, len)) {
      return -ENODEV;
   }

   memset(t, 0, sizeof(t));
   spi_message_init_with_transfers(&m, t, ARRAY_SIZE(t));

   /* Set up first few words for the SPI message */
   cmd[0] = (dev << FZM_CPLDSPI_DEV_SHIFT) | (read ? FZM_CPLDSPI_OP_READ : 0);

   switch (dev) {
   case FZM_CPLDSPI_CPLD:
      cmd[0] |= (addr >> FZM_CPLDSPI_CPLD_HIADDR_SHIFT) & FZM_CPLDSPI_CPLD_HIADDR_MASK;
      cmd[1] = addr & FZM_CPLDSPI_CPLD_LOADDR_MASK;
      t[0].len = 2;
      /* bursts are currently broken in the CPLD! */
      if (len != 1) {
         return -EMSGSIZE;
      }
      break;
   case FZM_CPLDSPI_FPGA_CONFIG:
      if (read) {
         return -EINVAL;
      }
      t[0].len = 1;
      break;
   default:
      return -EINVAL;
   }

   /* Populate SPI message */
   t[0].tx_buf = cmd;

   if (read)
      t[1].rx_buf = buf;
   else
      t[1].tx_buf = buf;

   t[1].len = len;

   ret = spi_sync(spidev, &m);
   if (ret) {
      dev_err(&spidev->dev, "spi_sync failed: %d\n", ret);
   }

   return ret;
}


int fzm_cpldspi_read(fzm_cpldspi_target_dev_t dev,
                     unsigned addr, void *buf, size_t len)
{
   return fzm_cpldspi_rw(dev, addr, buf, len, true);
}
EXPORT_SYMBOL(fzm_cpldspi_read);


int fzm_cpldspi_write(fzm_cpldspi_target_dev_t dev,
                      unsigned addr, const void *buf, size_t len)
{
   return fzm_cpldspi_rw(dev, addr, (void *)buf, len, false);
}
EXPORT_SYMBOL(fzm_cpldspi_write);


int fzm_cpldspi_read_locked(fzm_cpldspi_target_dev_t dev,
                            unsigned addr, void *buf, size_t len)
{
   int ret;

   fzm_cpldspi_lock(dev);
   ret = fzm_cpldspi_read(dev, addr, buf, len);
   fzm_cpldspi_unlock(dev);

   return ret;
}
EXPORT_SYMBOL(fzm_cpldspi_read_locked);


int fzm_cpldspi_write_locked(fzm_cpldspi_target_dev_t dev,
                             unsigned addr, const void *buf, size_t len)
{
   int ret;

   fzm_cpldspi_lock(dev);
   ret = fzm_cpldspi_write(dev, addr, buf, len);
   fzm_cpldspi_unlock(dev);

   return ret;
}
EXPORT_SYMBOL(fzm_cpldspi_write_locked);


static int fzm_cpldspi_probe(struct spi_device *spi)
{
   int err;

   if (WARN_ON(spidev))
      return -EEXIST;

   spidev = spi;

   err = cpld_ic_probe(&spi->dev);
   if (err)
      goto out;
   err = cpld_spi2c_probe(&spi->dev);
   if (err)
      goto out_ic;
   err = cpld_usr_probe(&spi->dev);
   if (err)
      goto out_spi2c;

   return 0;

out_spi2c:
   cpld_spi2c_remove(&spi->dev);
out_ic:
   cpld_ic_remove(&spi->dev);
out:
   return err;
}

static int fzm_cpldspi_remove(struct spi_device *spi)
{
   cpld_usr_remove(&spi->dev);
   cpld_spi2c_remove(&spi->dev);
   cpld_ic_remove(&spi->dev);

   return 0;
}

static const struct of_device_id fzm_cpldspi_id[] = {
   { .compatible = "nsn,fzm-cpld" },
   { }
};
MODULE_DEVICE_TABLE(of, fzm_cpldspi_id);

MODULE_ALIAS("spi:fzm-cpld");

static struct spi_driver fzm_cpldspi_driver = {
   .driver = {
      .name   = "fzm-cpld",
      .owner  = THIS_MODULE,
      .of_match_table = of_match_ptr(fzm_cpldspi_id),
   },
   .probe  = fzm_cpldspi_probe,
   .remove = fzm_cpldspi_remove,
};

module_spi_driver(fzm_cpldspi_driver);

MODULE_LICENSE("GPL v2");
MODULE_AUTHOR("Nokia Networks");
MODULE_DESCRIPTION("CPLD SPI driver for the NSN FZM");

