/*
 * File name: em_mod.c
 *
 * Description: OpenEM utility module to allow device user access and CMA
 *
 * Copyright (C) 2013 Texas Instruments, Incorporated
 * Author: Aurelien Jacquiot <a-jacquiot@ti.com>
 * 
 * Some code borrowed from Linux virtio_udma driver
 * Copyright (C) 2012, 2013 Texas Instruments Incorporated
 * Author: Cyril Chemparathy <cyril@ti.com>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation version 2.
 *
 * This program is distributed "as is" WITHOUT ANY WARRANTY of any kind,
 * whether express or implied; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */
#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/dma-mapping.h>
#include <linux/version.h>
#include <linux/wait-simple.h>
#include <linux/sched.h>
#include <linux/uaccess.h>
#include <linux/spinlock.h>
#include <linux/io.h>
#include <linux/hrtimer.h>
#include <linux/miscdevice.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/of.h>
#include <linux/ioctl.h>
#include <linux/module.h>
#include <linux/hwqueue.h>

#include <asm/cacheflush.h>
#include "linux/em_mod.h"

/* Number of total Linux OpenEM instances to be managed */
#define TI_EM_INSTANCE_NUM    2

#undef EM_MOD_DEBUG
#ifdef EM_MOD_DEBUG
#define DPRINTK(fmt, args...) printk(KERN_DEBUG "OpenEM module: [%s] " fmt, __FUNCTION__, ## args)
#else
#define DPRINTK(fmt, args...) no_printk(KERN_DEBUG "OpenEM module: [%s] " fmt, __FUNCTION__, ## args)
#endif

#define EPRINTK(fmt, args...) printk(KERN_CRIT "OpenEM module: [%s] ERROR " fmt, __FUNCTION__, ## args)

#define TI_EM_USE_INTERRUPT

struct em_device {
	struct miscdevice  misc;
	struct list_head   users;
	struct kref        refcount;
	struct swait_head  swait;
	atomic_t           event;
	int                intd;
	int                channel_base;
	unsigned long      irq_status;
	char               name[32];
};

struct em_user {
	int                instance;
	struct em_device  *em;
	struct list_head   node;
	struct file       *file;
	struct hwqueue    *queue;
	unsigned int       event_count;
};

#define from_misc(misc)	container_of(misc, struct em_device, misc)
#define to_misc(em)	(&(em)->misc)


static struct em_device *em_dev[TI_EM_INSTANCE_NUM];

static DEFINE_SPINLOCK(em_lock);

static void em_device_release(struct kref *kref)
{
	struct em_device *em;

	em = container_of(kref, struct em_device, refcount);

	WARN_ON(!list_empty(&em->users));

	DPRINTK("release em device instance 0x%x\n", (unsigned int) em);

	kfree(em);
}

static inline struct em_device *em_device_get(struct em_device *em)
{
	kref_get(&em->refcount);
	return em;
}

static inline void em_device_put(struct em_device *em)
{
	kref_put(&em->refcount, em_device_release);
}

static void em_add_user(struct em_device *em, struct em_user *user)
{
	spin_lock(&em_lock);
	list_add_tail(&user->node, &em->users);
	spin_unlock(&em_lock);
	em_device_get(em);
}

static void em_del_user(struct em_device *em, struct em_user *user)
{
	spin_lock(&em_lock);
	list_del(&user->node);
	spin_unlock(&em_lock);
	em_device_put(em);
}

#ifdef TI_EM_USE_INTERRUPT
/*
 * Manage incoming OpenEM interrupt-based event
 */
static int em_irq[TI_EM_INSTANCE_NUM];

#if !defined(CONFIG_OF_IRQ) || !defined(CONFIG_IRQ_DOMAIN) || !defined(CONFIG_ARM_GIC)
#error "only GIC based interrupt are supported"
#endif
#include <linux/interrupt.h>
#include <linux/irqdomain.h>
#include <linux/of_irq.h>

/* base of ARM hwirq for the various INTD */
#define QMSS_INTD1_HWIRQ_BASE            179
#define QMSS_INTD2_HWIRQ_BASE            229

#define GIC_IRQ_TYPE                     0xf01 /* pulsed interrupt */

/* QMSS INTD registers */
#define QMSS_INTD_EOI                    0x010
#define QMSS_INTD_COUNT                  0x300

/* QMSS INTD macros */
#define QMSS_INTD_COUNT_IRQ(c)           (QMSS_INTD_COUNT + ((c) << 2))
#define QMSS_INTD_CHANNEL(b, i)          ((b) + (i))
#define QMSS_INTD_CHANNEL_TO_HWIRQ(c, n) (em_hwirq_map[n] + (c))
#define QMSS_INTD_CHANNEL_TO_OFFSET(c)   (((c) & ~0x1f) >> 3)
#define QMSS_INTD_CHANNEL_TO_BITNUM(c)   ((c) & 0x1f)
#define QMSS_INTD_CHANNEL_TO_EOI(c)      ((c) + 2)

static const int em_hwirq_map[2] = { QMSS_INTD1_HWIRQ_BASE, QMSS_INTD2_HWIRQ_BASE };
static void __iomem *em_intd_regs[2];

/* GIC device tree match tables */
static const struct of_device_id irq_gic_match[] = {
        { .compatible = "arm,cortex-a15-gic", },
	{}
};

static int em_event_create_gic_virqs(int hwirq)
{
	struct device_node *np;
	const u32 intspec[3] = { 0, hwirq, GIC_IRQ_TYPE };

	np = of_find_matching_node(NULL, irq_gic_match);
	if (!np) {
		EPRINTK("Unable to find gic device node\n");
		return -1;
	}

	return irq_create_of_mapping(np, (const u32*) &intspec, 3);
}

/* Each register contains a count of the interrupts that have triggered and
 * not processed. In QMSS, this count saturates at 3. Reading the register
 * returns the count. Writing a non-zero value to the register subtracts
 * that value from the count. Writing a 0 clears the count.
 *
 * When the EOI is written, the interrupt will trigger again if the
 * corresponding Int Count Register is not 0
 */
static inline void em_event_intd_ack(int channel, int intd)
{
	void __iomem *intd_count_reg = em_intd_regs[intd] + QMSS_INTD_COUNT_IRQ(channel);

	/* Acknowledge all interrupts by clearing the COUNT */
	__raw_writel(0, intd_count_reg);
}

/* The EOI Register allows software to clear specific interrupts within the
 * INTD module. Unless interrupts have been cleared, they will not trigger
 * again. Each interrupt within QMSS is cleared by writing a specific 8-bit
 * value to the register. Writing to this register does not clear
 * corresponding bits in the Status Registers, nor does it clear interrupts
 * within the CorePac’s interrupt controller. When the EOI is written, the
 * interrupt will trigger again if the corresponding Int Count Register is
 * not 0, (this should not happen with firmware generated interrupts).
 */
static inline void em_event_intd_eoi(int channel, int intd)
{
	void __iomem *eoi = em_intd_regs[intd] + QMSS_INTD_EOI;

	__raw_writel(QMSS_INTD_CHANNEL_TO_EOI(channel), eoi);
}

static irqreturn_t em_event_interrupt_handler(int irq, void *_data)
{
	int instance = (int) _data;
	
	DPRINTK("Handler called for instance %d (irq %d)\n", instance, irq);

	/* increase logical event count*/
	atomic_inc(&em_dev[instance]->event);

	/* wake-up the waiting process if any*/
	swait_wake_all(&em_dev[instance]->swait);

	/* leave the interrupt disabled: do not write EOI */

	return IRQ_HANDLED;
}

static int em_event_wait(int instance, struct em_user *user)
{
	int ret;
	unsigned int event_count;
	int channel  = QMSS_INTD_CHANNEL(em_dev[instance]->channel_base, instance);
	int intd     = em_dev[instance]->intd;

	if (!em_irq[instance])
		return -ENODEV;

	/* Clear COUNT before checking the queue */
	em_event_intd_ack(channel, intd);

	/* although intd_ack and hwqueue_get_count both use uncacheable accesses,
	 * they are to different devices, and thus no ordering is guaranteed
	 * without a barrier */
	mb();

	ret = hwqueue_get_count(user->queue);
	if (ret < 0)
		return ret;
	else if (ret)
		return 0;

	/* write EOI here so that IRQ is enabled. COUNT will be non-zero if we
	 * have something in our queue now. */
	em_event_intd_eoi(channel, intd);

	DPRINTK("Waiting event for OpenEM instance %d\n", instance);

	ret = swait_event_interruptible(em_dev[instance]->swait,
	   (user->event_count != (event_count = atomic_read(&em_dev[instance]->event))));
	if (ret)
		goto out;

	user->event_count = event_count;

	DPRINTK("Receiving event for OpenEM instance %d, event = %u\n",
		instance, event_count);

out:
	return ret;
}

static int em_event_interrupt_init(void)
{
	em_intd_regs[0] = ioremap(0x02A0C000, 0x1000);
	if (em_intd_regs[0] == NULL) {
		EPRINTK("cannot remap QMMS INTD1\n");
		return -1;
	}
	
	em_intd_regs[1] = ioremap(0x02A0D000, 0x1000);
	if (em_intd_regs[1] == NULL) {
		EPRINTK("cannot remap QMMS INTD1\n");
		return -1;
	}
	return 0;
}

static int em_event_init(int instance, int channel_base, int intd, unsigned queue, struct em_user *user)
{
	int res;
	int hwirq;
	int channel = QMSS_INTD_CHANNEL(channel_base, instance);

	DPRINTK("creating mapping for instance %d, channel %d, intd %d\n",
		instance, channel, intd);

	if (em_irq[instance]) {
		free_irq(em_irq[instance], (void*) instance);
		em_irq[instance] = 0;
	}

	if (user->queue) {
		hwqueue_close(user->queue);
		user->queue = NULL;
	}

	em_event_intd_ack(channel, intd);

	/* Retrieve and register interrupts */
	hwirq = QMSS_INTD_CHANNEL_TO_HWIRQ(channel, intd);
	em_irq[instance] = em_event_create_gic_virqs(hwirq);

	DPRINTK("hwirq %d mapped to Linux virq %d\n", hwirq, em_irq[instance]);
	
	res = request_irq(em_irq[instance],
			  em_event_interrupt_handler,
			  IRQF_NO_THREAD,
			  em_dev[instance]->name,
			  (void*) instance);
	if (res < 0) {
		EPRINTK("cannot request irq %d: %d\n", em_irq[instance], res);
		em_irq[instance] = 0;
		return res;
	} 

	em_dev[instance]->channel_base = channel_base;
	em_dev[instance]->intd = intd;
	
	init_swait_head(&em_dev[instance]->swait);

	user->queue = hwqueue_open(em_dev[instance]->name, queue, O_RDONLY|O_NONBLOCK);
	if (IS_ERR(user->queue)) {
		res = PTR_ERR(user->queue);
		EPRINTK("cannot open hwqueue %u %d\n", queue, res);
		user->queue = NULL;
		goto out_irq;
	}

	return 0;

out_irq:
	free_irq(em_irq[instance], (void*) instance);
	em_irq[instance] = 0;
	return res;
}

static void em_event_release(int instance, struct em_user *user)
{
	DPRINTK("releasing instance %d, freeing irq %d\n", 
		instance, em_irq[instance]);
	
	if (em_irq[instance])
		free_irq(em_irq[instance], (void*) instance);
	
	if (em_dev[instance]->channel_base != -1) {
		int channel = 
			QMSS_INTD_CHANNEL(em_dev[instance]->channel_base, instance);
		em_event_intd_ack(channel, em_dev[instance]->intd);
	}
	
	em_irq[instance] = 0;
	em_dev[instance]->channel_base = -1;

	if (user && user->queue) {
		hwqueue_close(user->queue);
		user->queue = NULL;
	}
}
#else  /* TI_EM_USE_INTERRUPT */

static int em_event_init(int instance, int channel_base, int intd, struct em_user *user)
{
	return -ENOSYS;
}

static int em_event_wait(int instance, struct em_user *user)
{
	return -ENOSYS;
}

static void em_event_release(int instance, struct em_user *user) {}

#endif /* TI_EM_USE_INTERRUPT */

/*
 * File operation
 */
static long em_dev_fop_ioctl(struct file  *file,
			     unsigned int  cmd,
			     unsigned long arg)
{
	struct em_user *user = file->private_data;
	void __user    *argp = (void __user *)arg;

	switch(cmd) {
	case EM_IOC_WAITEVENTINIT:
	{	/* initialize event notification */
		struct ti_em_intd_irq_config irq_config;

		if (copy_from_user(&irq_config, argp, sizeof(irq_config)))
			return -EFAULT;

		return em_event_init(user->instance,
				     irq_config.channel_base,
				     irq_config.intd,
				     irq_config.queue,
				     user);
	}
	case EM_IOC_WAITEVENTFREE:
		/* free event notification */
		em_event_release(user->instance, user);
		break;
	case EM_IOC_WAITEVENT:
		/* wait for an incoming event */
		return em_event_wait(user->instance, user);
	default:
		return -ENOTTY;
	}

	return 0;
}

static int em_dev_fop_open(struct inode *inode, struct file *file)
{
	struct em_user *user;
	unsigned long instance;
	int res;

	res = kstrtoul(file->f_dentry->d_iname + (strlen(EM_MOD_DEVNAME)),
		       10, &instance);
	if (res != 0)
		return res;
	
	user = kzalloc(sizeof(*user), GFP_KERNEL);
	if (!user) {
		EPRINTK("failed to allocate user\n");
		return -ENOMEM;
	}
	
	user->em          = em_dev[instance];
	user->instance    = instance;
	user->file        = file;
	user->event_count = atomic_read(&em_dev[instance]->event);

	file->private_data = user;

	em_add_user(em_dev[instance], user);

	return 0;
}

static int em_dev_fop_release(struct inode *inode, struct file *file)
{
	struct em_user   *user = file->private_data;
	struct em_device *em   = user->em;

	em_event_release(user->instance, user);
	em_del_user(em, user);
	kfree(user);
	file->private_data = NULL;
	return 0;
}

static const struct file_operations em_dev_fops = {
	.owner             = THIS_MODULE,
	.open              = em_dev_fop_open,
	.release           = em_dev_fop_release,
	.unlocked_ioctl    = em_dev_fop_ioctl,
};

static int em_dev_register_device(void)
{
	struct miscdevice *misc;
	int res;
	int i;
	
#ifdef TI_EM_USE_INTERRUPT
	res = em_event_interrupt_init();
	if (res)
		return -ENOMEM;
#endif /* TI_EM_USE_INTERRUPT */
	
	for (i = 0; i < TI_EM_INSTANCE_NUM; i++) {
		em_dev[i] = kzalloc(sizeof(struct em_device), GFP_KERNEL);
		if (!em_dev[i]) {
			EPRINTK("cannot allocate device memory\n");
			return -ENOMEM;
		}

		sprintf((char *)&em_dev[i]->name, "%s%d", EM_MOD_DEVNAME, i);
	
		INIT_LIST_HEAD(&em_dev[i]->users);
		kref_init(&em_dev[i]->refcount);
		
		misc = to_misc(em_dev[i]);
		misc->minor = MISC_DYNAMIC_MINOR;
		misc->name  = em_dev[i]->name;
		misc->fops  = &em_dev_fops;

		res = misc_register(misc);
		if (res) {
			EPRINTK("could not register misc device for instance %d\n", i);
			return res;
		}

#ifdef TI_EM_USE_INTERRUPT
		em_irq[i] = 0;
#endif
	}

	return 0;
}

static void em_dev_unregister_device(void)
{
	struct miscdevice *misc;
	int i;

	for (i = 0; i < TI_EM_INSTANCE_NUM; i++) {
		int res;
		em_event_release(i, NULL);
		misc = to_misc(em_dev[i]);
		res = misc_deregister(misc);
		if (res)
			DPRINTK("cannot unregister instance %d (res = %d)\n", i, res);
		em_device_put(em_dev[i]);
	}
#ifdef TI_EM_USE_INTERRUPT
	iounmap(em_intd_regs[0]);
	iounmap(em_intd_regs[1]);
#endif
}

static int __init em_mod_init_module(void)
{
	/* register special device for EM */
	em_dev_register_device();

	return 0;
}

static void __exit em_mod_cleanup_module(void)
{
	em_dev_unregister_device();
}

module_init(em_mod_init_module);
module_exit(em_mod_cleanup_module);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Aurelien Jacquiot <a-jacquiot@ti.com>");
MODULE_DESCRIPTION("TI KeyStone OpenEM module");

