/*
 * Copyright 2014, 2015 Nokia Solutions and Networks
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
/**
 * @file em_wait.c
 *
 * @par REVISIONS
 * <pre>
 * Date       CoreID    SRID       Description
 * ---------- --------- ---------- ---------------------------------------------
 * 11/20/14   bwheele1             Initial Version
 * 01/06/15   bwheele1             Added support for infinite timeout/wait forever in WAIT_FOR ioctl
 * </pre>
 */

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/miscdevice.h>
#include <linux/err.h>
#include <linux/mm.h>
#include <linux/sched.h>
#include <linux/uaccess.h>
#include <linux/hwqueue.h>
#include <linux/ktime.h>
#include <linux/slab.h>
#include <linux/fs.h>

#include "linux/em_wait.h"

/* helpful command to use during testing: cat /sys/kernel/debug/hwqueues | grep -A 1 hwqueue-87 */

#define EM_WAIT_NAME "em_wait"

struct em_wait_hwq_info {
    struct hwqueue *hwq_hdl;
    wait_queue_head_t hwq_wait_q;
};

static int em_wait_open(struct inode *inode, struct file *file)
{
    struct em_wait_hwq_info *hwq_info;

    /* Allocate per-app info */
    hwq_info = kzalloc(sizeof(*hwq_info), GFP_KERNEL);
    if(hwq_info == NULL)
        return -ENOMEM;

    file->private_data = hwq_info;

    init_waitqueue_head(&hwq_info->hwq_wait_q);

    return 0;
}

static int em_wait_release(struct inode *inode, struct file *file)
{
    struct em_wait_hwq_info *hwq_info = file->private_data;

    if(hwq_info == NULL)
        return -EFAULT;

    if(hwq_info->hwq_hdl)
        hwqueue_close(hwq_info->hwq_hdl);

    kfree(hwq_info);

    return 0;
}

static void em_wait_hwq_int_callback(void *arg)
{
    struct em_wait_hwq_info *hwq_info = arg;

    /* Wake up any waiters */
    wake_up_all(&hwq_info->hwq_wait_q);
}

static int em_wait_ioctl_init(struct em_wait_hwq_info* const hwq_info, struct em_wait_ioctl_init __user *argp)
{
    struct em_wait_ioctl_init arg;
    char hwq_name[32];
    struct hwqueue *hwq_hdl;
    int err;

    if(copy_from_user(&arg, argp, sizeof(arg)))
        return -EFAULT;

    /*is hwq already configured?*/
    if(hwq_info->hwq_hdl)
        return -EBUSY;

    /*not important...doesn't affect what's displayed in /proc/interrupts*/
    snprintf(hwq_name, sizeof(hwq_name), "hwqueue-%u", arg.hwq_num);

    hwq_hdl = hwqueue_open(hwq_name, arg.hwq_num, O_RDONLY);
    if (IS_ERR(hwq_hdl)) {
        err = PTR_ERR(hwq_hdl);
        return err;
    }

    /*
    ** set callback function to use when interrupt happens for hwq
    **
    ** TI has no API to just set the callback function. hwqueue_set_notifier sets AND enables it via
    ** a call to hwqueue_enable_notifier.
    ** We don't want to enable interrupts just yet because that would just wait CPU cycles handling them.
    ** The only time we need to enable interrupts/the callback function is when the application calls the
    ** 'wait' ioctl.
    */
    hwq_hdl->notifier_fn = em_wait_hwq_int_callback;
    hwq_hdl->notifier_fn_arg = hwq_info;

    hwq_info->hwq_hdl = hwq_hdl;

    return 0;
}

static int em_wait_ioctl_wait_for(struct em_wait_hwq_info* const hwq_info, struct em_wait_ioctl_wait_for __user *argp)
{
    int err, rc;
    struct em_wait_ioctl_wait_for arg;

    if(copy_from_user(&arg, argp, sizeof(arg)))
        return -EFAULT;

    /*make sure hwq is configured*/
    if(hwq_info->hwq_hdl == NULL)
        return -ENXIO;

    /*optimization: if something already on the queue, we are done so return right away. No need to wait for the interrupt to happen immediately once interrupts are enabled*/
    if( (err = hwqueue_get_count(hwq_info->hwq_hdl)) < 0)
        return err;
    else if(err > 0)
        return 0;

    /*
    ** enable callback function to use when interrupt happens for hwq
    ** Note: we use enable instead of set_notifier to allow multiple waiters
    */
    if( (err = hwqueue_enable_notifier(hwq_info->hwq_hdl)) != 0)
        return err;

    /*put process to sleep until interrupt/msg is received or timeout occurs*/
    rc = wait_event_interruptible_hrtimeout(hwq_info->hwq_wait_q,
                                            hwqueue_get_count(hwq_info->hwq_hdl),
                                            arg.timeout_secs == EM_WAIT_IOCTL_INF_TIMEOUT?(ktime_t){ .tv64 = KTIME_MAX }:
                                                ktime_set(arg.timeout_secs, arg.timeout_nanosecs));

    /*don't restart ioctl automatically since the relative timeout won't be correct*/
    if(rc == -ERESTARTSYS)
        rc = -EINTR;

    /*disable the callback function set previously for the hwq*/
    WARN((err = hwqueue_disable_notifier(hwq_info->hwq_hdl)), "hwqueue_disable_notifier failed: %d\n", err);

    return rc;
}

static long em_wait_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
    void __user *argp = (void __user *)arg;
    struct em_wait_hwq_info *hwq_info = file->private_data;

    switch(cmd)
    {
        case EM_WAIT_INIT:
            return em_wait_ioctl_init(hwq_info, argp);
        case EM_WAIT_WAIT_FOR:
            return em_wait_ioctl_wait_for(hwq_info, argp);
        default:
            return -ENOTTY;
    }

    return 0;
}

static const struct file_operations em_wait_fops = {
    .owner = THIS_MODULE,
    .open = em_wait_open,
    .release = em_wait_release,
    .unlocked_ioctl = em_wait_ioctl,
};

static struct miscdevice em_wait_misc = {
    .minor = MISC_DYNAMIC_MINOR,
    .name = EM_WAIT_NAME,
    .fops = &em_wait_fops,
};

static int __init em_wait_init(void)
{
    return misc_register(&em_wait_misc);
}

static void __exit em_wait_exit(void)
{
    misc_deregister(&em_wait_misc);
}

module_init(em_wait_init);
module_exit(em_wait_exit);

MODULE_AUTHOR("Brian Wheeler");
MODULE_DESCRIPTION("Wait in hwqueue interrupt or timeout");
MODULE_LICENSE("GPL v2");
