#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/dma-mapping.h>
#include <linux/version.h>
#include <linux/wait-simple.h>
#include <linux/sched.h>
#include <linux/uaccess.h>
#include <linux/spinlock.h>
#include <linux/io.h>
#include <linux/miscdevice.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/of.h>
#include <linux/ioctl.h>
#include <linux/module.h>
#include <linux/idr.h>

#include "emtimer.h"

static struct kmem_cache *emtimer_cache;
static DEFINE_IDR(emtimer_ids);
static DEFINE_SPINLOCK(idr_lock);
static atomic_t usage_count = ATOMIC_INIT(1); /* set to 0 when cleaning up */
static DEFINE_MUTEX(cleanup_mutex);

static emtimer_t alloc_emtimer(struct emtimer **pt)
{
   struct emtimer *t;
   emtimer_t id = -1;
   int err;

   t = kmem_cache_zalloc(emtimer_cache, GFP_KERNEL);
   if (!t)
      return -ENOMEM;

   idr_preload(GFP_KERNEL);
   
   spin_lock(&idr_lock);
   id = idr_alloc(&emtimer_ids, t, 0, 0, GFP_NOWAIT);
   spin_unlock(&idr_lock);
   
   idr_preload_end();
   
   if (id < 0) {
      err = id;
      goto out_free;
   }

   *pt = t;

   pr_debug("allocated new timer %d: %p\n", id, t);
   
   return id;

out_free:
   kmem_cache_free(emtimer_cache, t);
   return err;
}

static struct emtimer* get_emtimer(emtimer_t id)
{
   struct emtimer *t;

   if (unlikely(id > INT_MAX))
      return NULL;

   rcu_read_lock();
   t = idr_find(&emtimer_ids, id);
   if (t) {
      if (!kref_get_unless_zero(&t->refcnt)) {
         t = NULL;
      }
   }
   rcu_read_unlock();
   
   return t;
}

static void emtimer_rcu_free(struct rcu_head *rcu)
{
   struct emtimer *t = container_of(rcu, struct emtimer, rcu);
   
   kmem_cache_free(emtimer_cache, t);
}

static void emtimer_kref_release(struct kref *kref)
{
   struct emtimer *t = container_of(kref, struct emtimer, refcnt);

   /* cancel synchronously to prevent expirations from rearming */
   hrtimer_cancel(&t->timer);

   /* use RCU to free in case someone was calling kref_get_unless_zero
    * between idr_find and rcu_read_unlock */
   call_rcu(&t->rcu, emtimer_rcu_free);
}

static void put_emtimer(struct emtimer *t)
{
   kref_put(&t->refcnt, emtimer_kref_release);
}

static void emtimer_remove(struct emtimer *t, emtimer_t id)
{
   /* cancel synchronously */
   /* Note that timer could get rearmed if this raced with a settime, but
    * that shouldn't happen since it is a use-after-free */
   hrtimer_cancel(&t->timer);

   /* remove from lookup table to prevent NEW lookups and allow reuse of id */
   spin_lock(&idr_lock);
   idr_remove(&emtimer_ids, id);
   spin_unlock(&idr_lock);

   /* free the original reference */
   put_emtimer(t);
}

static enum hrtimer_restart emtimer_hrtimer(struct hrtimer *hrtimer)
{
   struct emtimer *t = container_of(hrtimer, struct emtimer, timer);
   unsigned overruns;

   t->tx_count++;
   em_send(t);
   
   if (t->period.tv64) {
      /* check for overruns (1 is normal, 0 hasn't expired) */
      overruns = hrtimer_forward_now(&t->timer, t->period);
      if (overruns > 1) {
         overruns--;
         t->overrun_count += overruns;
         t->tx_count += overruns;
         while (overruns--) {
            em_send(t);
         }
      }
      return HRTIMER_RESTART;
   }else{
      return HRTIMER_NORESTART;
   }
}

static int emtimer_settime(struct emtimer *t, const struct itimerspec *its)
{
   int ret;
   unsigned overruns;

   ktime_t time = timespec_to_ktime(its->it_value);

   if (unlikely(!time.tv64)) {
      hrtimer_cancel(&t->timer);
   }else{
      t->period = timespec_to_ktime(its->it_interval);

      /* ensure that any interval is >= 50us so we don't get stuck in IRQ */
      if (ktime_to_ns(t->period) && ktime_to_ns(t->period) < 50000) {
         return -EDEADLK;
      }

      /* allow timer to expire up to 1us later to optimize IRQs */
      ret = hrtimer_start_range_ns(&t->timer, time, 1000, HRTIMER_MODE_ABS);
      /* PREEMPT_RT doesn't put expired irqsafe timers on list */
      if (ret == -ETIME) {
         if (t->period.tv64) {
            /* periodic */
            int maxretries = 1;
            
            do {
               overruns = hrtimer_forward_now(&t->timer, t->period);
               
               /* limit the maximum number of overruns at creation to guard
                * against OOB data and prevent buffer depletion */
               if (overruns > 16) {
                  return -ETIME;
               }
               
               t->overrun_count += overruns;
               t->tx_count += overruns;
               while (overruns--) {
                  em_send(t);
               }
            } while (-ETIME == hrtimer_restart(&t->timer) && maxretries--);
            
            if (maxretries < 0) {
               return -ETIME;
            }
         }else{
            /* one-shot */
            t->overrun_count++;
            t->tx_count++;
            return em_send(t);
         }
      }
   }
   
   return 0;
}

static long emtimer_config_ioc(struct emtimer_config __user *argp)
{
   struct emtimer_config arg;

   if (copy_from_user(&arg, argp, sizeof(arg)))
      return -EFAULT;
   
   //TODO
   //destroy all timers or prevent reconfig

   return em_send_config(&arg);
}

static long emtimer_create_ioc(struct emtimer_create __user *argp)
{
   struct emtimer_create arg;
   emtimer_t id;
   struct emtimer *t;
   int err;

   if (copy_from_user(&arg, argp, sizeof(arg)))
      return -EFAULT;
   
   id = alloc_emtimer(&t);
   if (id < 0)
      return id;

   kref_init(&t->refcnt);
   hrtimer_init(&t->timer, CLOCK_MONOTONIC, HRTIMER_MODE_ABS);
   t->timer.irqsafe = 1; /* for PREEMPT_RT */
   t->timer.function = emtimer_hrtimer;

   t->queue = arg.queue;

   err = em_send_create(t, &arg);
   if (err)
      goto out;

   emtimer_settime(t, &arg.its);

   if ((err = put_user(id, &argp->timerid)))
      goto out;
   
   return 0;

out:
   emtimer_remove(t, id);
   
   return err;
}

static long emtimer_settime_ioc(const struct emtimer_settime __user *argp)
{
   struct emtimer_settime arg;
   struct emtimer *t;
   int err;

   if (copy_from_user(&arg, argp, sizeof(arg)))
      return -EFAULT;
   
   t = get_emtimer(arg.timerid);
   if (!t) {
      return -ENOENT;
   }

   err = emtimer_settime(t, &arg.its);
   
   put_emtimer(t);
   
   return err;
}

static long emtimer_destroy_ioc(struct emtimer_destroy __user *argp)
{
   struct emtimer_destroy arg;
   struct emtimer *t;

   if (copy_from_user(&arg, argp, sizeof(arg)))
      return -EFAULT;

   t = get_emtimer(arg.timerid);
   if (!t) {
      return -ENOENT;
   }

   emtimer_remove(t, arg.timerid);
   /* there are no NEW users at this point */

   arg.tx_count = t->tx_count;
   arg.overrun_count = t->overrun_count;

   put_emtimer(t);
   
   if (copy_to_user(argp, &arg, sizeof(arg)))
      return -EFAULT;

   return 0;
}


static long emtimer_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
	void __user *argp = (void __user *)arg;

	switch (cmd) {
	case EMTIMER_CONFIG:
	   return emtimer_config_ioc(argp);
	case EMTIMER_CREATE:
	   return emtimer_create_ioc(argp);
	case EMTIMER_SETTIME:
	   return emtimer_settime_ioc(argp);
	case EMTIMER_DESTROY:
	   return emtimer_destroy_ioc(argp);
	default:
		return -ENOTTY;
	}

	return 0;
}

static int emtimer_open(struct inode *inode, struct file *file)
{
   if (unlikely(!atomic_inc_not_zero(&usage_count))) {
      /* wait for cleanup to complete */
      mutex_lock(&cleanup_mutex);
      atomic_inc(&usage_count);
      mutex_unlock(&cleanup_mutex);
   }

   return 0;
}

static int destroy_emtimers_cb(int id, void *p, void *arg)
{
   struct emtimer *t = p;
   unsigned *count = arg;

   *count += 1;
   
   hrtimer_cancel(&t->timer);
   /* no reads can come in since the cdev is not open */
   kmem_cache_free(emtimer_cache, p);
   
   return 0;
}

static void emtimer_cleanup(void)
{
   unsigned count = 0;

   spin_lock(&idr_lock);
   idr_for_each(&emtimer_ids, destroy_emtimers_cb, &count);
   idr_destroy(&emtimer_ids);
   spin_unlock(&idr_lock);

   if (count)
      pr_warn("emtimer: destroyed %u outstanding timers after last close\n", count);
}

static int emtimer_release(struct inode *inode, struct file *file)
{
   /* destroy all timers on last close */
   if (atomic_dec_return(&usage_count) == 1) {
      mutex_lock(&cleanup_mutex);
      /* try to clean up, unless someone else incremented in the meantime */
      if (atomic_cmpxchg(&usage_count, 1, 0) == 1) {
         emtimer_cleanup();
         atomic_set(&usage_count, 1);
      }
      mutex_unlock(&cleanup_mutex);
   }
   
   return 0;
}

static const struct file_operations emtimer_fops = {
	.owner             = THIS_MODULE,
	.unlocked_ioctl    = emtimer_ioctl,
	.open              = emtimer_open,
	.release           = emtimer_release,
};

static struct miscdevice emtimer_misc = {
   .minor = MISC_DYNAMIC_MINOR,
   .name = "emtimer",
   .fops = &emtimer_fops,
};

static int __init emtimer_init_module(void)
{
   int err;

   err = em_send_init();
   if (err)
      return err;

   emtimer_cache = kmem_cache_create("emtimer_cache", sizeof(struct emtimer),
      0, 0, NULL);
   if (!emtimer_cache) {
      err = -ENOMEM;
      goto out_em;
   }

   idr_init(&emtimer_ids);

   err = misc_register(&emtimer_misc);
   if (err)
      goto out_idr;
   
   return 0;

out_idr:
   idr_destroy(&emtimer_ids);

   kmem_cache_destroy(emtimer_cache);
   
out_em:
   em_send_exit();

   return err;
}


static void __exit emtimer_cleanup_module(void)
{
   misc_deregister(&emtimer_misc);
   
   /* wait for any RCU callbacks to complete */
   rcu_barrier();
   
   kmem_cache_destroy(emtimer_cache);
   
   em_send_exit();
}

module_init(emtimer_init_module);
module_exit(emtimer_cleanup_module);

MODULE_LICENSE("GPL v2");
MODULE_AUTHOR("Matt Cowell <matt.cowell@nsn.com>");
MODULE_DESCRIPTION("Event Machine timer driver");
