#ifndef __FZM_BOARD_TYPE_H__
#define __FZM_BOARD_TYPE_H__

#include <linux/of.h>
#include <stdbool.h>

#define BOARD_TYPE_OF_NODE_PATH "/board-identity"
#define BOARD_TYPE_PROP_NAME "board-type"

#define WIFI_SUPPORT_OF_NODE_PATH "/board-identity/user-defined-a"
#define WIFI_SUPPORT_PROP_NAME "wifisupport"

#define BH3_TYPE_OF_NODE_PATH "/board-identity/user-defined-a"
#define BH3_TYPE_PROP_NAME "backhaul-3-type"

#define BH1_TYPE_OF_NODE_PATH "/board-identity/user-defined-a"
#define BH1_TYPE_PROP_NAME "backhaul-1-type"

// values defined below should match u-boot/board/nsn/argon_k2/eid.h

#define BT_MASTER_SOLO 0
#define BT_MASTER_WITH_SLAVE 1
#define BT_SLAVE 2
#define BT_FZP 3
#define BT_4X4 4
#define BT_FZP_DUAL_BAND 5
#define BT_HP 6
#define BT_DBO_1G 7
#define BT_DBO 8

// values are copied from u-boot/board/nsn/argon_k2/board_type.h

#define RUCKUS_WIFI_SUPPORTED 0x1
#define CAMBRIDGE_WIFI_SUPPORTED 0x2
#define BH1_TYPE_COPPER 0x1
#define BH1_TYPE_SFP_FIBER 0x2
#define BH3_TYPE_SFP_FIBER 0x1
#define BH3_TYPE_SWITCH_MULTIPORT 0x2
#define BH3_TYPE_COPPER 0x6

typedef enum {
   MAC_FIBER,
   MAC_PHY,
   MAC_MAC,
   MAC_NOT_CONNECTED
} sgmii_type_t;

static inline int get_board_type(void) {
    struct device_node* pNode;
    u32 value;
    int ret = -1;

    pNode = of_find_node_by_path(BOARD_TYPE_OF_NODE_PATH);
    if (pNode == NULL)
        return ret;

    ret = of_property_read_u32(pNode, BOARD_TYPE_PROP_NAME, &value);

    of_node_put(pNode);

    return (ret?: value);
}

static inline bool is_fzp_board(void) {
    return (get_board_type() == BT_FZP);
}

static inline bool is_4x4_board(void) {
    return (get_board_type() == BT_4X4);
}

static inline bool is_fzp_db_board(void) {
    return (get_board_type() == BT_FZP_DUAL_BAND);
}

static inline bool is_fzm_dbo_1g_board(void) {
    return (get_board_type() == BT_DBO_1G);
}

static inline bool is_fzm_dbo_board(void) {
    return (get_board_type() == BT_DBO);
}

static inline bool is_wifi_supported(void)
{
    struct device_node* pNode = NULL;
    u32 value;
    int ret;

    pNode = of_find_node_by_path(WIFI_SUPPORT_OF_NODE_PATH);
    if (pNode == NULL)
        return false;

    ret = of_property_read_u32(pNode, WIFI_SUPPORT_PROP_NAME, &value);

    of_node_put(pNode);

    return ret ? false : ((value == RUCKUS_WIFI_SUPPORTED) || (value == CAMBRIDGE_WIFI_SUPPORTED));
}

static inline int get_backhaul1_type(void)
{
    struct device_node* pNode = NULL;
    int ret = -1;
    u32 value;

    pNode = of_find_node_by_path(BH1_TYPE_OF_NODE_PATH);
    if (pNode == NULL)
        return ret;

    ret = of_property_read_u32(pNode, BH1_TYPE_PROP_NAME, &value);

    of_node_put(pNode);

    return (ret?: value);
}

static inline int get_backhaul3_type(void)
{
    struct device_node* pNode = NULL;
    int ret = -1;
    u32 value;

    /* this is a temporary hack until the new EID format goes in */
    if(is_4x4_board())
        return BH3_TYPE_SFP_FIBER;

    pNode = of_find_node_by_path(BH3_TYPE_OF_NODE_PATH);
    if (pNode == NULL)
        return ret;

    ret = of_property_read_u32(pNode, BH3_TYPE_PROP_NAME, &value);

    of_node_put(pNode);

    return (ret?: value);
}

static inline sgmii_type_t get_sgmii_type(unsigned sgmii_nr) {
   int bh1_type = get_backhaul1_type();
   int bh3_type = get_backhaul3_type();
   int bt = get_board_type();

   /* All SGMII ports are MAC_MAC on DBO 10G */
   if (BT_DBO == bt)
      return MAC_MAC;

   switch (sgmii_nr)
   {
      case 0: 
         /* On Argon HP and DBO_1G, this is BH3 copper */
         if ((BT_HP == bt) || (BT_DBO_1G == bt))
            return MAC_PHY;

         /* For non-Argon HP boards, if BH1 type is not copper, something is wrong */
         if (bh1_type != BH1_TYPE_COPPER)
           return MAC_NOT_CONNECTED;

         return MAC_PHY;
         break;

      case 1: 
         /* All BH types are copper */
         return MAC_PHY;
         break;

      case 2:
         /* Is either BH1 fiber or not connected */
         if (BH1_TYPE_SFP_FIBER == bh1_type)
           return MAC_FIBER;
         else
           return MAC_NOT_CONNECTED;
         break;
 
      case 3:
         /* Special case for Argon HP - return fiber */
         if (BT_HP == bt)
            return MAC_FIBER;

         /* Special case for Argon DBO_1G - return mac-mac */
         if (BT_DBO_1G == bt)
            return MAC_MAC;

         /* Otherwise, return BH3 type */
         switch (bh3_type)
         {
            case BH3_TYPE_SFP_FIBER: 
               return MAC_FIBER;
               break;
            case BH3_TYPE_COPPER: 
               return MAC_PHY;
               break;
            case BH3_TYPE_SWITCH_MULTIPORT:
               return MAC_MAC;
               break;
            default:
               return MAC_NOT_CONNECTED;
               break;
         }
         break;

      default:
         return MAC_NOT_CONNECTED;
         break;
   }
}
#endif /* __FZM_BOARD_TYPE_H__ */
