/* Copyright 2014 Nokia Solutions and Networks */

/*******************************MODULE_HEADER*************************
 *
 *  Module Name:    fzm_fpga_eis_gpspi.h
 *
 *  Description:    This file contains defines, register definitions
 *                  interfacing with the FPGA via K2 SPI1-CS0
 *
 *
 *  mm/dd/yyyy CR Num  Description of Change                  Engineer
 *  ---------- ------- -------------------------------------- -----------
 *  06/06/2012 1476403 Created file                           C. Powe
 *  04/15/2013         Port to FZM                            yongxu
 *
 *********************************************************************/

#ifndef FZM_FPGA_EIS_GPSPI_H
#define FZM_FPGA_EIS_GPSPI_H

/* ========================================================================= */
/* FPGA EIS DETAILS                                                          */
/* ========================================================================= */

/*
 ** FPGA configuration is intended to be via the K2 SPI buses.
 ** The K2 can access FPGA registers using SPI1-CS0
 **
 ** Each access shall consist minimally of a single 8-bit word.
 ** A full command SPI transaction will be used to initiate a FPGA action.
 ** A single 8-bit word will be used during master poling of FPGA write
 ** transactions to reduce chatter.
 ** Read poling should use the full command word
 **
 **  Byte 0:
 ** +------+------+------+------+------+------+------+------+
 ** |  MT3 |  MT2 |  MT1 |  MT0 | DID3 | DID2 | DID1 | DID0 |
 ** +------+------+------+------+------+------+------+------+
 **
 ** DID3-0: 4 bits indicating Device ID
 **         0 - for FPGA
 **         1 - for ADI Catalina
 **         2 - for PLL ?
 ** MT3-0: 4 bits message type
 **        0x0 Polling
 **        0x1 Generic Read
 **        0x2 Generic Write
 **        0x3 Generic Read Burst
 **        0x4 Generic Write Burst
 **        0x5 Device Read
 **        0x6 Device Write
 **
 **  Byte1 - Byte4:
 ** +-----+-----+-----+------+-----+-----+-----+-----+
 ** |                  Payload0 MSB                  |
 ** +-----+-----+-----+------+-----+-----+-----+-----+
 ** |                Payload0 Nibble1                |
 ** +-----+-----+-----+------+-----+-----+-----+-----+
 ** |                Payload0 Nibble2                |
 ** +-----+-----+-----+------+-----+-----+-----+-----+
 ** |                  Payload0 LSB                  |
 ** +-----+-----+-----+------+-----+-----+-----+-----+
 **
 ** MT dependent:
 **     - Polling, set to 0
 **     - Generic/Device Read/Write: address
 **     - Generic Read Burst, first 32 bits of data to be read from the
 **       address auto-increment to the next 32-bit boundary from the
 **       previous non burst request
 **     - Generic Write Burst, first 32 bits of data to be writen to the
 **       address auto-increment to the next 32-bit boundary from the
 **       previous non burst request
 **
 **  Byte5 - Byte8:
 ** +-----+-----+-----+------+-----+-----+-----+-----+
 ** |                  Payload1 MSB                  |
 ** +-----+-----+-----+------+-----+-----+-----+-----+
 ** |                Payload1 Nibble1                |
 ** +-----+-----+-----+------+-----+-----+-----+-----+
 ** |                Payload1 Nibble2                |
 ** +-----+-----+-----+------+-----+-----+-----+-----+
 ** |                  Payload1 LSB                  |
 ** +-----+-----+-----+------+-----+-----+-----+-----+
 **
 ** MT dependent:
 **     - Polling, set to 0
 **     - Generic/Device Read: first 32 bits data to be read from the
 **                            address in payload0
 **     - Generic/Device Write: first 32 bits data to be written to the
 **                            address in payload0
 **     - Generic Read Burst, second 32 bits of data to be read from the
 **       address auto-increment to the next 32-bit boundary from the
 **       address used in payload0
 **     - Generic Write Burst, second 32 bits of data to be written to the
 **       address auto-increment to the next 32-bit boundary from the
**       address used in payload0
**
**
*/

//TODO: move to devicetree
#define FPGA_SIZE 0x100000000ULL

#define FPGA_SPI_MT_SHIFT            (4) //JAP:FIX HLD
#define FPGA_SPI_MT_MASK             (0xF)

#define FPGA_SPI_DEVID_SHIFT         (0)  //JAP:FIX HLD
#define FPGA_SPI_DEVID_MASK          (0xF)

#define FPGA_SPI_PLD_MSB_SHIFT    (24)
#define FPGA_SPI_PLD_MSB_MASK     (0xFF)
#define FPGA_SPI_PLD_NIB1_SHIFT   (16)
#define FPGA_SPI_PLD_NIB1_MASK    (0xFF)
#define FPGA_SPI_PLD_NIB2_SHIFT   (8)
#define FPGA_SPI_PLD_NIB2_MASK    (0xFF)
#define FPGA_SPI_PLD_LSB_SHIFT    (0)
#define FPGA_SPI_PLD_LSB_MASK     (0xFF)

#define SIEN1R       0x00000100
#define SIEN2R       0x00000104
#define SIRQ1R       0x00000110
#define SIRQ2R       0x00000114
#define RP3_INT_BIT     0
#define REFAB_INT_BIT      12

#define RP3_IRQ         0x0014001c
#define RP3_IRQ_MASK    0x00140020

#define RP3_IRQ_REAL_TIME_STATUS        0x00140028
#define SSTA1R                          0x00000120

#define RP3_LMTSMODECHANGE_BIT  24
#define REFABON_BIT             12

typedef enum FPGA_SPI_MT_e {
   FPGA_SPI_MT_POLL           = 0x0,
   FPGA_SPI_MT_GEN_READ       = 0x1,
   FPGA_SPI_MT_GEN_WRITE      = 0x2,
   FPGA_SPI_MT_GEN_READ_BRST  = 0x3,
   FPGA_SPI_MT_GEN_WRITE_BRST = 0x4,
   FPGA_SPI_MT_DEV_READ       = 0x5,
   FPGA_SPI_MT_DEV_WRITE      = 0x6
} FPGA_SPI_MT_t;

typedef enum FPGA_SPI_DEVID_e {
   FPGA_SPI_DEVID_N_A      = 0x0,
   FPGA_SPI_DEVID_ADI_CAT  = 0x1, /* Catalina on all FZMs except FWQA; Clemente01 on FWQA */
   FPGA_SPI_DEVID_CLE01    = 0x1,
   FPGA_SPI_DEVID_CLE23    = 0x2, /* Clemente23 on FWQA boards, unused on other FZM variants */
   FPGA_SPI_DEVID_ADC      = 0x3, /* ADC on FWQA boards, unused on other FZM variants */
   FPGA_SPI_MAX_DEVS
} FPGA_SPI_DEVID_t;

typedef enum {
   FPGA_EIS_SPI_OK,
   FPGA_EIS_SPI_INVALID_PARAM,
   FPGA_EIS_SPI_FAILURE,
   FPGA_EIS_SPI_INTERNAL_FAILURE,
   FPGA_EIS_SPI_NUM_STATUS
} FPGA_EIS_SPI_STATUS_E;

#endif /* FZM_FPGA_EIS_RFSPI_H */

